# libs/fitting.py

import numpy as np
import matplotlib.pyplot as plt
import math
from scipy.optimize import least_squares
from numpy.random import default_rng
from scipy.stats import chi2 as stats_chi2
import sys


def straight_line(params, xdata):
    """
    Function for a straight line
    - params   array of function parameters
    - xdata    array of x data points

    """

    f = params[0] + params[1] * xdata
    return f


def straight_line_diff(params, xdata):
    """
    Differential of function for a straight line
    - params   array of function parameters
    - xdata    array of x data points

    """

    df = params[1]
    return df


def polynomial(params, xdata):
    """
    Function for a ploynomial of order `len(nparams)`
    - params   array of function parameters
    - xdata    array of x data points

    """

    f = 0

    npol = len(params)
    for i in range(npol):
        f += params[i] * xdata**i

    return f


# <!-- Polynomial Fit -- START -->


def polynomial_diff(params, xdata):
    """
    Differential of function for a polynomial of order `len(nparams)`
    - params   array of function parameters
    - xdata    array of x data points
    """

    df = 0

    npol = len(params)
    for i in range(1, npol):
        df += params[i] * i * xdata**(i - 1)

    return df


# <!-- Polynomial Fit -- END -->


def minimise(params, xdata, ydata, xerr, yerr):
    """
    Calculation to minimise to find the best fit using chi2: difference between each y point and its
    prediction  by the function, divided by the sum in quadrature of the error on y, both from the
    y error and from the related error in x.
    - params   array of function parameters
    - xdata    array of x data points
    - ydata    array of y data points
    - xerr     array of x data errors
    - yerr     array of y data errors
    - func     function describing data
    - diff     differential of function
    """

    residuals = (ydata - straight_line(params, xdata)) / (
        np.sqrt(yerr**2 + straight_line_diff(params, xdata)**2 * xerr**2))

    return residuals


# Minimise Poly -- START
def minimise_poly(params, xdata, ydata, xerr, yerr):
    """
    Calculation to minimise to find the best fit using chi2: difference between each y point and its
    prediction  by the function, divided by the sum in quadrature of the error on y, both from the
    y error and from the related error in x.
    - params   array of function parameters
    - xdata    array of x data points
    - ydata    array of y data points
    - xerr     array of x data errors
    - yerr     array of y data errors
    - func     function describing data
    - diff     differential of function
    """

    residuals = (ydata - polynomial(params, xdata)) / (
        np.sqrt(yerr**2 + polynomial_diff(params, xdata)**2 * xerr**2))

    return residuals


def fit(xdata,
        ydata,
        xerror,
        yerror,
        init_params=None,
        xlabel="X",
        ylabel="Y",
        title=None,
        fname=None):
    """
    Function to perform least-squares fit for a straight line.
    - xdata         array of x data points
    - ydata         array of y data points
    - xerr          array of x data errors
    - yerr          array of y data errors
    - init_params   array of function parameters
    - xlabel        string of the x axis
    - ylabel        string of the y axis
    - title         string of the title of graph
    """

    # Run fit
    if init_params is None:
        init_params = [0.0, 10.0]

    if title is None:
        title = ""
    else:
        title = title

    if fname is None:
        fname = "unknown"
    else:
        fname = fname

    result = least_squares(minimise,
                           init_params,
                           args=(xdata, ydata, xerror, yerror))

    # Check fit succeeds
    if not result.success or result.status < 1:
        print("ERROR: Fit failed with message {}".format(result.message))
        print("Please check the data and inital parameter estimates")
        return 0, 0
    else:
        print("Fit succeeded")

    # Get fitted parameters
    final_params = result.x
    c = final_params[0]
    m = final_params[1]
    nparams = len(final_params)

    # Calculate chi2
    chi2_array = result.fun**2
    chi2 = sum(chi2_array)
    npoints = len(xdata)
    reduced_chi2 = chi2 / (npoints - nparams)
    chi2_prob = stats_chi2.sf(chi2, (npoints - nparams))

    # Print chi2
    np.set_printoptions(precision=3)
    print("\n=== Fit quality ===")
    print("chisq per point = \n", chi2_array)
    print(
        "chisq = {:7.5g}, ndf = {}, chisq/NDF = {:7.5g}, chisq prob = {:7.5g}\n"
        .format(chi2, npoints - nparams, reduced_chi2, chi2_prob))

    if reduced_chi2 < 0.25 or reduced_chi2 > 4:
        print(
            "WARNING: chi2/ndf suspiciously small or large.  Please check the data and initial parameter estimates"
        )

    if chi2_prob < 0.05:
        print(
            "WARNING: chi2 probability for given degrees of freedom less than 0.05 .  Please check the data and "
            "initial parameter estimates")

        # Calculate errors
    jacobian = result.jac
    jacobian2 = np.dot(jacobian.T, jacobian)
    determinant = np.linalg.det(jacobian2)

    if determinant < 1E-32:
        print(
            f"Matrix singular (determinant = {determinant}, error calculation failed."
        )
        param_errors = np.zeros(nparams)
    else:
        covariance = np.linalg.inv(jacobian2)
        param_errors = np.sqrt(covariance.diagonal())

    print("=== Fitted parameters ===")
    print("c = {:7.5g} +- {:7.5g}".format(final_params[0], param_errors[0]))
    print("m = {:7.5g} +- {:7.5g}".format(final_params[1], param_errors[1]))

    # Calculate fitted function values
    yfit = straight_line(final_params, xdata)

    # Visualise result
    fig = plt.figure(figsize=(5, 4))
    plt.title(title)
    plt.xlabel(xlabel, fontsize=13)
    plt.ylabel(ylabel, fontsize=13)
    plt.grid(color='grey', linestyle="--")

    plt.errorbar(xdata,
                 ydata,
                 xerr=xerror,
                 yerr=yerror,
                 fmt='k',
                 marker=".",
                 linestyle='',
                 label="Data")
    plt.plot(xdata, yfit, color='r', linestyle='-', label="Fit")

    plt.legend(loc=0, fontsize=16)

    text = "c: {:7.5g} +- {:7.5g}\n".format(final_params[0], param_errors[0])
    text += "m: {:7.5g} +- {:7.5g}\n".format(final_params[1], param_errors[1])
    plt.text(0.95,
             0,
             text,
             transform=fig.axes[0].transAxes,
             ha="right",
             va="bottom",
             fontsize=12)

    plt.show()
    # plt.savefig(fname)

    # Return arrays of parameters and associated errors
    return final_params, param_errors


# <!-- Fit Data as a linear graph - END -->
# Fit Poly


def fit_poly(xdata,
             ydata,
             xerror,
             yerror,
             init_params=None,
             xlabel="X",
             ylabel="Y",
             title=None,
             fname=None):
    """
    Function to perform least-squares fit for a straight line.
    - xdata         array of x data points
    - ydata         array of y data points
    - xerr          array of x data errors
    - yerr          array of y data errors
    - init_params   array of function parameters
    - xlabel        string of the x axis
    - ylabel        string of the y axis
    - title         string of the title of graph
    """

    # Run fit
    if init_params is None:
        init_params = [0.0, 10.0]

    if title is None:
        title = ""
    else:
        title = title

    if fname is None:
        fname = "unknown"
    else:
        fname = fname

    result = least_squares(minimise_poly,
                           init_params,
                           args=(xdata, ydata, xerror, yerror))

    # Check fit succeeds
    if not result.success or result.status < 1:
        print("ERROR: Fit failed with message {}".format(result.message))
        print("Please check the data and inital parameter estimates")
        return 0, 0
    else:
        print("Fit succeeded")

    # Get fitted parameters
    final_params = result.x
    c = final_params[0]
    m = final_params[1]
    nparams = len(final_params)

    # Calculate chi2
    chi2_array = result.fun**2
    chi2 = sum(chi2_array)
    npoints = len(xdata)
    reduced_chi2 = chi2 / (npoints - nparams)
    chi2_prob = stats_chi2.sf(chi2, (npoints - nparams))

    # Print chi2
    np.set_printoptions(precision=3)
    print("\n=== Fit quality ===")
    print("chisq per point = \n", chi2_array)
    print(
        "chisq = {:7.5g}, ndf = {}, chisq/NDF = {:7.5g}, chisq prob = {:7.5g}\n"
        .format(chi2, npoints - nparams, reduced_chi2, chi2_prob))

    if reduced_chi2 < 0.25 or reduced_chi2 > 4:
        print(
            "WARNING: chi2/ndf suspiciously small or large.  Please check the data and initial parameter estimates"
        )

    if chi2_prob < 0.05:
        print(
            "WARNING: chi2 probability for given degrees of freedom less than 0.05 .  Please check the data and "
            "initial parameter estimates")

        # Calculate errors
    jacobian = result.jac
    jacobian2 = np.dot(jacobian.T, jacobian)
    determinant = np.linalg.det(jacobian2)

    if determinant < 1E-32:
        print(
            f"Matrix singular (determinant = {determinant}, error calculation failed."
        )
        param_errors = np.zeros(nparams)
    else:
        covariance = np.linalg.inv(jacobian2)
        param_errors = np.sqrt(covariance.diagonal())
    finalText = ""

    if nparams == 2:
        print("=== Fitted parameters ===")
        print("Intercept (c) = {:7.5g} +- {:7.5g}".format(
            final_params[0], param_errors[0]))
        finalText += "c: {:7.5g} +- {:7.5g}\n".format(final_params[0],
                                                      param_errors[0])
        print("gradient (m) = {:7.5g} +- {:7.5g}".format(
            final_params[1], param_errors[1]))
        finalText += "m: {:7.5g} +- {:7.5g}\n".format(final_params[1],
                                                      param_errors[1])
    else:
        print("(Parameters in order of increasing power of x)")
        for x in range(nparams):
            print("param {} = {:7.5g} +- {:7.5g}".format(
                x, final_params[x], param_errors[x]))
            finalText += "p{}: {:7.5g} +- {:7.5g}\n".format(
                x, final_params[x], param_errors[x])

    # Calculate fitted function values
    yfit = straight_line(final_params, xdata)

    # Visualise result
    fig = plt.figure(figsize=(5, 4))
    plt.title(title)
    plt.xlabel(xlabel, fontsize=13)
    plt.ylabel(ylabel, fontsize=13)
    plt.grid(color='grey', linestyle="--")

    plt.errorbar(xdata,
                 ydata,
                 xerr=xerror,
                 yerr=yerror,
                 fmt='k',
                 marker=".",
                 linestyle='',
                 label="Data")
    plt.plot(xdata, yfit, color='r', linestyle='-', label="Fit")

    plt.legend(loc=0, fontsize=16)

    text = "c: {:7.5g} +- {:7.5g}\n".format(final_params[0], param_errors[0])
    text += "m: {:7.5g} +- {:7.5g}\n".format(final_params[1], param_errors[1])
    plt.text(0.95,
             0,
             finalText,
             transform=fig.axes[0].transAxes,
             ha="right",
             va="bottom",
             fontsize=12)

    plt.show()
    # plt.savefig(fname)

    # Return arrays of parameters and associated errors
    return final_params, param_errors


# <!-- Fit data as a histogram - START -->


def plotHistogram(array,
                  title,
                  xLabel,
                  yLabel,
                  binBot=0,
                  binTop=1000,
                  binNumber=10):
    if binBot == 0 and binTop == 1000:
        binBot = min(array)
        binTop = max(array)
    else:
        binBot = binBot
        binTop = binTop

    bins, binWidth = np.linspace(binBot, binTop, binNumber + 1, retstep=True)

    # Calculate mean and standard deviation
    nEvents = len(array)
    mu = np.mean(array)  # calculate arithmetic mean of numbers in array
    sigma = np.std(
        array)  # calculate standard deviation (error on single value)
    muError = sigma / np.sqrt(
        nEvents)  # calculate error of mean from sigma/sqrt(n)

    # Construct a figure with a title and axis labels
    plt.figure(figsize=(7, 5))
    plt.title(title, fontsize=14)
    plt.xlabel(xLabel)
    plt.ylabel(yLabel)

    # Make a histogram and display it
    plt.hist(array, bins=bins, color='b')
    plt.grid(color='grey')
    plt.show()

    print("Histogram bins start at", binBot, "finish at", binTop)
    print("Number of bins is", binNumber, "and width of bins is", binWidth)

    return mu, sigma, muError


# <!-- Fit data as a histogram - END -->
# <!-- Consistency Check Function - START -->


def consistencycheck(m1, m2, mErr1, mErr2):
    leftSide = np.abs(m1 - m2)
    rightSide = 3 * ((math.sqrt((mErr1)**2 + (mErr2)**2)))

    if leftSide <= rightSide:
        print("The results are consistent ",
              f"{round(leftSide, 4)} <= {round(rightSide, 4)}")
    else:
        print("The results are inconsistent",
              f"{round(leftSide, 4)} > {round(rightSide, 4)}")


# <!-- Consistency Check Function - END -->

# <!-- Polynomial Fit -- END -->


# <!-- Weighted Mean Function -- START -->
def weightedMean(data, errors):
    for i in errors:
        weights = 1 / i**2

    numer = np.sum(np.array(data) * np.array(weights))
    denom = np.sum(np.array(weights))

    newdata = np.average(data)
    newerror = 1 / math.sqrt(denom)

    return newdata, newerror


# <!-- Weighted Mean Function -- END -->


def generate_toy_data(xarray, slope, intercept):
    """
    Generate toy data, randomly Poisson fluctuated around y = slope * xarray + intercept.
    Take a simple Gaussian error i.e. sqrt(N) on each y point.
    """

    model = intercept + slope * xarray

    rng = default_rng(3)
    yarray = rng.poisson(model)
    yerr = np.sqrt(yarray)

    return yarray, yerr
